package org.greenstone.gatherer.metadata;


import java.io.*;
import java.util.*;
import org.greenstone.gatherer.util.Utility;
import org.greenstone.gatherer.util.XMLTools;
import org.w3c.dom.*;


/** This class represents one doc.xml file */
public class DocXMLFile
    extends File
{
    static final private String ARCHIVE_ELEMENT = "Archive";
    static final private String DESCRIPTION_ELEMENT = "Description";
    static final private String METADATA_ELEMENT = "Metadata";
    static final private String SECTION_ELEMENT = "Section";

    private ArrayList files_in_doc_xml_file = new ArrayList();


    public DocXMLFile(String doc_xml_file_path)
    {
	super(doc_xml_file_path);
    }


    public ArrayList getMetadataExtractedFromFile(File file)
    {
	// Build up a list of metadata extracted from this file
	ArrayList metadata_values = new ArrayList();

	String file_relative_path = file.getAbsolutePath();
	int import_index = file_relative_path.indexOf("import");
	if (import_index != -1) {
	    file_relative_path = file_relative_path.substring(import_index + "import".length() + 1);
	}

	// Check whether this doc.xml file contains extracted metadata for the specified file
	boolean contains_extracted_metadata_for_file = false;
	for (int i = 0; i < files_in_doc_xml_file.size(); i++) {
	    if (file_relative_path.equals(files_in_doc_xml_file.get(i))) {
		contains_extracted_metadata_for_file = true;
		break;
	    }
	}

	// ...it doesn't
	if (!contains_extracted_metadata_for_file) {
	    return metadata_values;
	}

	// Parse the doc.xml file
	System.err.println("Applicable doc.xml file: " + this);
	Document document = XMLTools.parseXMLFile(this);
	if (document == null) {
	    System.err.println("Error: Could not parse doc.xml file " + getAbsolutePath());
	    return metadata_values;
	}

	MetadataSet extracted_metadata_set = MetadataSetManager.getMetadataSet(MetadataSetManager.EXTRACTED_METADATA_NAMESPACE);

	// Read all the Archive elements in the file
	NodeList archive_elements_nodelist = document.getElementsByTagName(ARCHIVE_ELEMENT);
	for (int i = 0; i < archive_elements_nodelist.getLength(); i++) {
	    Element current_archive_element = (Element) archive_elements_nodelist.item(i);

	    // Read the child Section elements of the archive (but not all descendants)
	    ArrayList child_section_elements = XMLTools.getChildElementsByTagName(current_archive_element, SECTION_ELEMENT);
	    for (int j = 0; j < child_section_elements.size(); j++) {
		Element current_section_element = (Element) child_section_elements.get(j);

		// Read the Description elements of this section only (not child sections as well)
		ArrayList child_description_elements = XMLTools.getChildElementsByTagName(current_section_element, DESCRIPTION_ELEMENT);
		for (int k = 0; k < child_description_elements.size(); k++) {
		    Element current_description_element = (Element) child_description_elements.get(k);

		    // Read all the Metadata elements in this description element
		    NodeList metadata_elements_nodelist = current_description_element.getElementsByTagName(METADATA_ELEMENT);
		    for (int l = 0; l < metadata_elements_nodelist.getLength(); l++) {
			Element current_metadata_element = (Element) metadata_elements_nodelist.item(l);
			String metadata_element_name_full = current_metadata_element.getAttribute("name");

			// If the metadata has a namespace it isn't extracted metadata, so we're not interested
			String metadata_set_namespace = MetadataTools.getMetadataSetNamespace(metadata_element_name_full);
			if (!metadata_set_namespace.equals("")) {
			    continue;
			}

			// Extracted metadata!
			String metadata_element_name = metadata_element_name_full;

			// Ignore lower-case metadata elements (gsdlsourcefilename, gsdlassocfile etc.)
			//   and those starting with '/' (/srclink)
			char first_character = metadata_element_name.charAt(0);
			if (Character.isLowerCase(first_character) || first_character == '/') {
			    continue;
			}

			MetadataElement metadata_element = extracted_metadata_set.getMetadataElement(metadata_element_name);

			// Value trees are not stored for extracted metadata, so create a new value tree node now
			String current_metadata_element_value = XMLTools.getElementTextValue(current_metadata_element);
			metadata_element.addMetadataValue(current_metadata_element_value);
			MetadataValueTreeNode metadata_value_tree_node = metadata_element.getMetadataValueTreeNode(current_metadata_element_value);

			// Add the new metadata value to the list
			MetadataValue metadata_value = new MetadataValue(metadata_element, metadata_value_tree_node);
			metadata_values.add(metadata_value);
		    }
		}
	    }
	}

	return metadata_values;
    }


    /**
     * Every doc.xml file must be skimmed when a collection is opened, for two reasons:
     *   - To build a mapping from source file to its corresponding doc.xml file
     *   - To get a complete list of all extracted metadata elements
     */
    public void skimFile()
    {
	MetadataSet extracted_metadata_set = MetadataSetManager.getMetadataSet(MetadataSetManager.EXTRACTED_METADATA_NAMESPACE);

	// Skim the doc.xml file as quickly as possible (don't parse as XML), looking at the Metadata elements
	System.err.println("Skimming doc.xml file " + this + "...");
	try {
	    BufferedReader buffered_reader = new BufferedReader(new FileReader(this));
	    String line = null;
	    while ((line = buffered_reader.readLine()) != null) {
		// This line doesn't contain a metadata element
		if (line.indexOf("<Metadata ") == -1) {
		    continue;
		}

		// Extract the metadata element name
		int name_index = line.indexOf(" name=\"") + " name=\"".length();
		String metadata_element_name_full = line.substring(name_index, line.indexOf("\"", name_index));

		// If the metadata has a namespace it isn't extracted metadata, so we're not interested
		String metadata_set_namespace = MetadataTools.getMetadataSetNamespace(metadata_element_name_full);
		if (!metadata_set_namespace.equals("")) {
		    continue;
		}

		// Extracted metadata!
		String metadata_element_name = metadata_element_name_full;

		// Note which file this doc.xml is for
		if (metadata_element_name.equals("gsdlsourcefilename")) {
		    // Extract the gsdlsourcefilename element value
		    int value_index = line.indexOf(">", name_index) + ">".length();
		    String gsdlsourcefilename_value = line.substring(value_index, line.indexOf("<", value_index));

		    // We're only interested in the path relative to the import folder
		    int import_index = gsdlsourcefilename_value.indexOf("import");
		    if (import_index != -1) {
			gsdlsourcefilename_value = gsdlsourcefilename_value.substring(import_index + "import".length());

			boolean is_unix_path = gsdlsourcefilename_value.startsWith("/");
			gsdlsourcefilename_value = gsdlsourcefilename_value.substring(1);

			// Make sure the path matches the OS that is running
			if (is_unix_path && Utility.isWindows()) {
			    // Convert path from Unix to Windows
			    gsdlsourcefilename_value = gsdlsourcefilename_value.replaceAll("\\/", "\\\\");
			}
			else if (!is_unix_path && !Utility.isWindows()) {
			    // Convert path from Windows to Unix
			    gsdlsourcefilename_value = gsdlsourcefilename_value.replaceAll("\\\\", "/");
			}

			// Remember this for quick access later
			files_in_doc_xml_file.add(gsdlsourcefilename_value);
		    }
		    else {
			// We don't really know what is going on...
			System.err.println("Warning: Could not understand gsdlsourcefilename " + gsdlsourcefilename_value);
		    }
		}

		// Ignore lower-case metadata elements (gsdlsourcefilename, gsdlassocfile etc.)
		//   and those starting with '/' (/srclink)
		char first_character = metadata_element_name.charAt(0);
		if (Character.isLowerCase(first_character) || first_character == '/') {
		    continue;
		}

		MetadataElement metadata_element = extracted_metadata_set.getMetadataElement(metadata_element_name);
		if (metadata_element == null) {
		    // This element isn't defined in ex.mds, so create it for this session
		    System.err.println("Extracted metadata element not defined: " + metadata_element_name);
		    extracted_metadata_set.addMetadataElementForThisSession(metadata_element_name);
		}
	    }
	}
	catch (Exception ex) {
	    System.err.println("Exception: " + ex);
	    ex.printStackTrace();
	}
    }
}
